# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dispy.modules.dictwrapper import DictWrapper
from dispy.types.variable import Snowflake, Timestamp
from typing import List, Dict, Any

from dispy.types.role import Role
from dispy.types.emoji import Emoji
from dispy.types.stickers import Sticker
from dispy.types.user import User, AvatarDecorationData

class WelcomeScreenChannel(DictWrapper):
    channel_id: Snowflake
    description: str
    emoji_id: Snowflake
    emoji_name: str


class WelcomeScreen(DictWrapper):
    description: str
    welcome_channels: List[WelcomeScreenChannel]


class Member(DictWrapper):
    user: User
    nick: str
    avatar: str
    banner: str
    roles: List[Snowflake]
    joined_at: Timestamp
    premium_since: Timestamp
    deaf: bool
    mute: bool
    flags: int
    pending: bool
    permissions: str
    communication_disabled_until: Timestamp
    avatar_decoration_data: AvatarDecorationData
    unusual_dm_activity_until: Timestamp

class Guild(DictWrapper):
    id: Snowflake
    name: str
    icon: str
    icon_hash: str
    splash: str
    discovery_splash: str
    owner: bool
    owner_id: Snowflake
    permissions: str
    region: str
    afk_channel_id: Snowflake
    afk_timeout: int
    widget_enabled: bool
    widget_channel_id: Snowflake
    verification_level: int
    default_message_notifications: int
    explicit_content_filter: int
    roles: List[Role]
    emojis: List[Emoji]
    features: List[str]
    mfa_level: int
    application_id: Snowflake
    system_channel_id: Snowflake
    system_channel_flags: int
    rules_channel_id: Snowflake
    max_presences: int
    max_members: int
    vanity_url_code: str
    description: str
    banner: str
    premium_tier: int
    premium_subscription_count: int
    preferred_locale: str
    public_updates_channel_id: Snowflake
    max_video_channel_users: int
    max_stage_video_channel_users: int
    approximate_member_count: int
    approximate_presence_count: int
    welcome_screen: WelcomeScreen
    nsfw_level: int
    locale: str
    stickers: List[Sticker]
    premium_progress_bar_enabled: bool
    safety_alerts_channel_id: Snowflake


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dispy.modules.dictwrapper import DictWrapper
from dispy.types.variable import Snowflake, Timestamp
from typing import List, Dict, Any

from dispy.types.emoji import Emoji
from dispy.types.guild import Member

class ReactionCountDetails(DictWrapper):
    burst: int
    normal: int

class Reaction(DictWrapper):
    count: int
    count_details: ReactionCountDetails
    me: bool
    me_burst: bool
    emoji: Emoji
    burst_colors: List[Any]

class ReactionAdd(DictWrapper):
    user_id: Snowflake
    channel_id: Snowflake
    message_id: Snowflake
    guild_id: Snowflake
    member: Member
    emoji: Emoji
    type: int
    message_author_id: Snowflake
    burst: bool
    burst_colors: List[Any]

class ReactionRemove(DictWrapper):
    user_id: Snowflake
    channel_id: Snowflake
    message_id: Snowflake
    guild_id: Snowflake
    emoji: Emoji
    type: int
    burst: bool

class ReactionRemoveAll(DictWrapper):
    channel_id: Snowflake
    message_id: Snowflake
    guild_id: Snowflake

class ReactionRemoveEmoji(DictWrapper):
    channel_id: Snowflake
    message_id: Snowflake
    guild_id: Snowflake
    emoji: Emoji


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# This file is used to attribute arguments to a eventname
# For example, when a message is send, we give the message and the author object separatly.

from dispy.types.message import Message
from dispy.types.interaction import Interaction
from dispy.types.user import User
from dispy.types.reaction import ReactionAdd, ReactionRemove, ReactionRemoveAll, ReactionRemoveEmoji
from dispy.types.variable import Null
from typing import Any, get_type_hints

from dispy.modules import dict_to_obj
class _eventargs:
    """
    🚫 Don't use it if you don't know what you're doing.
    """
    def __init__(self,intents,error):
        self.intents = intents
        self.error = error
        
    def set(self,eventname,api, **kwargs):
        if eventname == 'READY':
            return {}
        
        # Messages
        elif eventname in ['MESSAGE_CREATE','MESSAGE_UPDATE','DIRECT_MESSAGE_CREATE','DIRECT_MESSAGE_UPDATE']:
            msg = Message(**kwargs, _api=api)
            user = msg.author
            return {'msg': msg,'user': user}
        
        # Reactions
        elif eventname in ['MESSAGE_REACTION_ADD','DIRECT_MESSAGE_REACTION_ADD']:
            return {'reaction': ReactionAdd(**kwargs, _api=api)}
        elif eventname in ['MESSAGE_REACTION_REMOVE','DIRECT_MESSAGE_REACTION_REMOVE']:
            return {'reaction': ReactionRemove(**kwargs, _api=api)}
        elif eventname in ['MESSAGE_REACTION_REMOVE_ALL','DIRECT_MESSAGE_REACTION_REMOVE_ALL']:
            return {'reaction': ReactionRemoveAll(**kwargs, _api=api)}
        elif eventname in ['MESSAGE_REACTION_REMOVE_EMOJI','DIRECT_MESSAGE_REACTION_REMOVE_EMOJI']:
            return {'reaction': ReactionRemoveEmoji(**kwargs, _api=api)}
        
        # Interactions
        elif eventname == 'INTERACTION_CREATE':
            return {'int': Interaction(**kwargs, _api=api)}
        else:
            return {'args': dict_to_obj(kwargs)}
        
    def get(self,eventname):
        if eventname == 'READY':
            return {}
        
        # Messages
        elif eventname in ['MESSAGE_CREATE','MESSAGE_UPDATE','DIRECT_MESSAGE_CREATE','DIRECT_MESSAGE_UPDATE']:
            return {'msg': Message, 'user': User}
        
        # Reactions
        elif eventname in ['MESSAGE_REACTION_ADD','DIRECT_MESSAGE_REACTION_ADD']:
            return {'reaction': ReactionAdd}
        elif eventname in ['MESSAGE_REACTION_REMOVE','DIRECT_MESSAGE_REACTION_REMOVE']:
            return {'reaction': ReactionRemove}
        elif eventname in ['MESSAGE_REACTION_REMOVE_ALL','DIRECT_MESSAGE_REACTION_REMOVE_ALL']:
            return {'reaction': ReactionRemoveAll}
        elif eventname in ['MESSAGE_REACTION_REMOVE_EMOJI','DIRECT_MESSAGE_REACTION_REMOVE_EMOJI']:
            return {'reaction': ReactionRemoveEmoji}
        
        # Interactions
        elif eventname == 'INTERACTION_CREATE':
            return {'int': Interaction}
        else:
            return {'args': Any}

    def check_function(self,function,eventname):
        code = function.__code__
        function_arguments = list(code.co_varnames[:code.co_argcount])
        type_hints = get_type_hints(function)
        current_types = {arg: type_hints.get(arg,Null) for arg in function_arguments}
        needed_type = self.get(eventname)
        stringcode = ", ".join([f"{key}: {value.__name__}" for key, value in needed_type.items()])

        if len(needed_type) == 0:
            if not len(current_types) == 0:
                self.error.summon('noargs',eventname=eventname,function_name=function.__name__)
        else:
            for name, ntypes in needed_type.items():
                # Check for missing arguments
                if name not in current_types:
                    self.error.summon('missing_args', function_name=function.__name__, name=name, arguments=stringcode)

                # Check for extra arguments (we don't want those shit)
                for key in current_types:
                    if key not in needed_type:
                        self.error.summon('extra_args', function_name=function.__name__, name=key, arguments=stringcode)

                # Check for types
                input_type = current_types[name]
                if ntypes is not Any and not issubclass(input_type,ntypes):
                    self.error.summon('invalidtype_args', name=name, needed_type=ntypes.__name__, function_name=function.__name__, arguments=stringcode)
        return True


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dispy.modules.dictwrapper import DictWrapper
from dispy.types.variable import Snowflake, Timestamp
from typing import List, Dict, Any

from dispy.types.permissions import Overwrite
from dispy.types.user import User
from dispy.types.thread import ThreadMetadata, ThreadMember, ForumTag, DefaultReaction


class ChannelMention(DictWrapper):
    id: Snowflake
    guild_id: Snowflake
    type: int
    name: str


class Channel(DictWrapper):
    id: Snowflake
    type: int
    guild_id: Snowflake
    position: int
    permission_overwrites: List[Overwrite]
    name: str
    topic: str
    nsfw: bool
    last_message_id: Snowflake
    bitrate: int
    user_limit: int
    rate_limit_per_user: int
    recipients: List[User]
    icon: str
    owner_id: Snowflake
    application_id: Snowflake
    managed: bool
    parent_id: Snowflake
    last_pin_timestamp: Timestamp
    rtc_region: str
    video_quality_mode: int
    message_count: int
    member_count: int
    thread_metadata: ThreadMetadata
    member: ThreadMember
    default_auto_archive_duration: int
    permissions: str
    flags: int
    total_message_sent: int
    available_tags: List[ForumTag]
    applied_tags: List[Snowflake]
    default_reaction_emoji: DefaultReaction
    default_thread_rate_limit_per_user: int
    default_sort_order: int
    default_forum_layout: int


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dispy.modules.dictwrapper import DictWrapper
from dispy.types.variable import Snowflake, Timestamp
from typing import List, Dict, Any

class RoleSubscriptionData(DictWrapper):
    role_subscription_listing_id: Snowflake
    tier_name: str
    total_months_subscribed: int
    is_renewal: bool

# This function is shit, see the discord wiki to see it: https://discord.com/developers/docs/topics/permissions#role-object-role-tags-structure
# I will need to change bool to True if null is present, else to False, who is the dumbass at discord that coded that api shit.
# \/ \/ \/ \/

class RoleTags(DictWrapper):
    bot_id: Snowflake
    integration_id: Snowflake
    premium_subscriber: Any # Will not work
    subscription_listing_id: Snowflake
    available_for_purchase: Any # Will not work
    guild_connections: Any # Will not work

class Role(DictWrapper):
    id: Snowflake
    name: str
    color: int
    hoist: bool
    icon: str
    unicode_emoji: str
    position: int
    permissions: str
    managed: bool
    mentionable: bool
    tags: RoleTags
    flags: int


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dispy.modules.dictwrapper import DictWrapper
from dispy.types.variable import Snowflake, Timestamp
from typing import List, Dict, Any

from dispy.types.user import User
from dispy.types.guild import Guild
from dispy.types.team import Team

class InstallParams(DictWrapper):
    scopes: List[str]
    permissions: str

class Application(DictWrapper):
    id: Snowflake
    name: str
    icon: str
    description: str
    rpc_origins: List[str]
    bot_public: bool
    bot_require_code_grant: bool
    bot: User
    terms_of_service_url: str
    privacy_policy_url: str
    owner: User
    verify_key: str
    team: Team
    guild_id: Snowflake
    guild: Guild
    primary_sku_id: Snowflake
    slug: str
    cover_image: str
    flags: int
    approximate_guild_count: int
    approximate_user_install_count: int
    redirect_uris: List[str]
    interactions_endpoint_url: str
    role_connections_verification_url: str
    event_webhooks_url: str
    event_webhooks_status: int
    event_webhooks_types: List[str]
    tags: List[str]
    install_params: InstallParams
    integration_types_config: Dict[Any, Any]
    custom_install_url: str


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

"""
Dispy is a light-weight discord API library.
It is recommended to import it with `from dispy import *`
"""
# Internal
from dispy.modules.intents import *
from dispy.modules import *
from dispy.types.user import User
from dispy.modules.rest_api import __internal__ as restapi
from dispy.modules.error import error
from dispy.modules.eventargs import _eventargs
from dispy.data import data
# External
from typing import Callable, Literal
from concurrent.futures import ThreadPoolExecutor
import aiohttp # Need to be installed (with websocket_client)
import json
import threading
import time
from warnings import deprecated
import asyncio

# 888888ba  oo                               
# 88    `8b ``                               
# 88     88 dP .d8888b.    88d888b. dP    dP 
# 88     88 88 Y8ooooo.    88'  `88 88    88 
# 88    .8P 88       88 ,, 88.  .88 88.  .88 
# 8888888P  dP `88888P' 88 88Y888P' `8888P88 
#                          88            .88 
#                          dP        d8888P  

# Developed by ✯James French✯ with ❤ and hopes x)
# Licensed with GPLv3

class Bot(restapi): # <- this shit has taken me hours
    #--------------------------------------------------------------------------------------#
    #                                      Bot Setup                                       #
    #--------------------------------------------------------------------------------------# 
    def __init__(self,token=None):
        """
        Define your bot.

        See related page on the [wiki](https://jamesfrench.gitbook.io/dispy).
        """
        self.user: User = None
        self.status = 0

        self._token = token
        self._error = error()
        self._heartbeat_interval = None
        self._handlers = []
        self._session: aiohttp.ClientSession = None
        self._ws = None
        self._api = restapi(self._token,self._error)
        self._data = data()
        self._eventargs = _eventargs(self._data.intents,self._error)
        self._loop = asyncio.new_event_loop()
        self._executor = ThreadPoolExecutor()
        self._tasks = []
        threading.Thread(target=self._run_loop, daemon=True).start()
    def _run_loop(self):
        asyncio.set_event_loop(self._loop)
        self._loop.run_forever() # no_traceback

    def __getattr__(self, name):
        return getattr(self._api, name)

    def config(self,token=None):
        if self.status != 0: return None
        if token != None: self._token = token

    #--------------------------------------------------------------------------------------#
    #                                    Internal Code                                     #
    #--------------------------------------------------------------------------------------#
    # MAIN: Run the bot and get events
    async def _main(self):
        """
        🚫 Don't use it if you don't know what you're doing.
        """
        async for msg in self._ws:
            data = json.loads(msg.data)
    
            if data['op'] == 10: # Identification and heartbeat set
                self._heartbeat_interval = data['d']['heartbeat_interval'] / 1000
                asyncio.create_task(self._heartbeat())
    
                await self._identify()
            else: # Events
                if data['t'] != None:
                    if data['t'] == "READY":
                        self.user = User(**data['d']['user'])
                    asyncio.create_task(self._sendevent(data['t'],data['d']))

    # Used to call functions when a event is dispatched
    async def _sendevent(self, eventname, args):
        """
        🚫 Don't use it if you don't know what you're doing.
        """
        once_remove = []
        tasks = []
    
        for key, handler in self._handlers:
            if key == eventname:
                if eventname in self._data.intents.direct_intents_opposed:
                    if handler['is_direct'] and 'guild_id' in args:
                        continue
                    if not handler['is_direct'] and 'guild_id' not in args:
                        continue
    
                arguments = self._eventargs.set(eventname, self._api, **args)
                task = asyncio.to_thread(handler['function'], **arguments)
                tasks.append(task)
    
                if handler.get('once', False):
                    once_remove.append((key, handler))
    
        if tasks:
            await asyncio.gather(*tasks)
        for key, handler in once_remove:
            self._handlers.remove((key, handler))

    # Make the bot online
    async def _identify(self):
        """
        🚫 Don't use it if you don't know what you're doing.
        """
        payload = {
            'op': 2,
            'd': {
                'token': self._token,
                'intents': self._intents(),
                'properties': {
                    'os': 'linux',
                    'browser': 'dispy-lib',
                    'device': 'dispy-lib'
                }
            }
        }
        await self._ws.send_json(payload)

    # Send heartbeat to discord to keep the bot alive
    async def _heartbeat(self):
        """
        🚫 Don't use it if you don't know what you're doing.
        """
        while self.status == 1: 
            await asyncio.sleep(self._heartbeat_interval)
            if self.status != 1:
                break
            heartbeat_payload = {
                "op": 1,
                "d": 'null'
            }
            await self._ws.send_json(heartbeat_payload)

    # Used to calculate intents
    def _intents(self):
        """
        🚫 Don't use it if you don't know what you're doing.
        """
        events = set()
        ids = []
        for key, handler in self._handlers:
            event_name = f'DIRECT_{key}' if handler['is_direct'] else key
            events.add(event_name)  # Add event name to the set

        ids = list(set([id for event in events if (intent_found := self._data.intents.get_intents(event)) for id in intent_found]))
        return sum(1 << int(id) for id in ids)

        
    # Start the bot
    async def _start(self):
        self._session = aiohttp.ClientSession()
        try:
            async with self._session.ws_connect('wss://gateway.discord.gg/?v=10&encoding=json') as ws:
                self._ws = ws
                await self._main()
        finally:
            await self._session.close()
        return None

    #--------------------------------------------------------------------------------------#
    #                                     Bot Control                                      #
    #--------------------------------------------------------------------------------------#
    def run(self) -> None:
        """
        Start your bot and make it online.
        Make your bot capable of receiving events and sending events.
        """
        if self.status != 0: self._error.summon('bot_is_already_running')
        self.status = 1
        
        asyncio.run(self._start()) # no_traceback
        time.sleep(2)

    def stop(self) -> None:
        """
        Shutdown the bot. (EXPERIMENTAL)
        """
        async def _stop():
            if self._ws:
                await self._ws.close(code=1000)
            if self._session:
                await self._session.close()
            self.status = 0

        asyncio.run_coroutine_threadsafe(_stop(),loop=self._loop)

    def custom_request(self,function,path,payload=None):
        async def asynchronous():
            return await self._api.__request__(function=function,path=path,payload=payload)
        
        future_result = asyncio.run_coroutine_threadsafe(asynchronous(), self._loop)
        return future_result.result(timeout=7)

    def _debug(self):
        pass

    #--------------------------------------------------------------------------------------#
    #                                    Event Handler                                     #
    #--------------------------------------------------------------------------------------#

    global _eventliteral
    _eventliteral = Literal["GUILD_BAN_ADD", "MESSAGE_UPDATE", "GUILD_CREATE", "DIRECT_MESSAGE_REACTION_REMOVE_ALL", "GUILD_ROLE_CREATE", "GUILD_SCHEDULED_EVENT_CREATE", "GUILD_DELETE", "GUILD_SCHEDULED_EVENT_UPDATE", "ALL", "MESSAGE_REACTION_REMOVE_ALL", "GUILD_MEMBER_REMOVE", "INVITE_DELETE", "STAGE_INSTANCE_CREATE", "DIRECT_CHANNEL_PINS_UPDATE", "CHANNEL_DELETE", "GUILD_ROLE_UPDATE", "DIRECT_MESSAGE_CREATE", "DIRECT_MESSAGE_REACTION_REMOVE_EMOJI", "MESSAGE_DELETE_BULK", "THREAD_UPDATE", "MESSAGE_POLL_VOTE_REMOVE", "GUILD_SOUNDBOARD_SOUND_DELETE", "VOICE_STATE_UPDATE", "GUILD_INTEGRATIONS_UPDATE", "USER_UPDATE", "GUILD_ROLE_DELETE", "MESSAGE_REACTION_REMOVE", "DIRECT_MESSAGE_UPDATE", "MESSAGE_DELETE", "GUILD_SCHEDULED_EVENT_DELETE", "THREAD_MEMBER_UPDATE", "PRESENCE_UPDATE", "INTEGRATION_UPDATE", "GUILD_SOUNDBOARD_SOUND_CREATE", "WEBHOOKS_UPDATE", "GUILD_AUDIT_LOG_ENTRY_CREATE", "AUTO_MODERATION_RULE_DELETE", "READY", "AUTO_MODERATION_RULE_UPDATE", "THREAD_CREATE", "DIRECT_MESSAGE_POLL_VOTE_ADD", "RESUMED", "INTEGRATION_DELETE", "GUILD_UPDATE", "THREAD_DELETE", "GUILD_SOUNDBOARD_SOUNDS_UPDATE", "INVITE_CREATE", "MESSAGE_POLL_VOTE_ADD", "DIRECT_MESSAGE_REACTION_REMOVE", "CHANNEL_PINS_UPDATE", "MESSAGE_REACTION_REMOVE_EMOJI", "GUILD_MEMBER_UPDATE", "GUILD_MEMBER_ADD", "CHANNEL_CREATE", "VOICE_CHANNEL_EFFECT_SEND", "MESSAGE_REACTION_ADD", "GUILD_SCHEDULED_EVENT_USER_REMOVE", "GUILD_EMOJIS_UPDATE", "INTERACTION_CREATE", "DIRECT_MESSAGE_POLL_VOTE_REMOVE", "CHANNEL_UPDATE", "GUILD_BAN_REMOVE", "DIRECT_MESSAGE_DELETE", "VOICE_SERVER_UPDATE", "DIRECT_TYPING_START", "AUTO_MODERATION_RULE_CREATE", "GUILD_STICKERS_UPDATE", "MESSAGE_CREATE", "STAGE_INSTANCE_UPDATE", "THREAD_LIST_SYNC", "GUILD_SCHEDULED_EVENT_USER_ADD", "TYPING_START", "GUILD_SOUNDBOARD_SOUND_UPDATE", "INTEGRATION_CREATE", "THREAD_MEMBERS_UPDATE", "DIRECT_MESSAGE_REACTION_ADD", "AUTO_MODERATION_ACTION_EXECUTION", "STAGE_INSTANCE_DELETE"]

    def on(self, eventname: _eventliteral = None, function: Callable = None, *, once: bool = False) -> None:
        """
        Add a function to call when a specific event is dispatched.
        """
        def decorator(fn):
            if self.status != 0: self._error.summon('bot_is_running')

            if eventname is None: event_name = fn.__name__.upper()
            else: event_name = eventname.upper()
            if event_name in self._data.intents.intents:
                if self._eventargs.check_function(fn,event_name): # no_traceback
                    is_direct = event_name in self._data.intents.direct_intents
                    event_name = event_name[7:] if is_direct else event_name

                    self._handlers.append((event_name,{
                        "function": fn,
                        "is_direct": is_direct,
                        "once": once,
                    }))
            else:
                self._error.summon("event_invalid",event=event_name.upper())
        if function is not None: return decorator(function)
        else: return decorator

    def once(self, eventname: _eventliteral = None, function: Callable = None, *, once: bool = True) -> None:
        """
        Add a function to call when a specific event is dispatched once.
        """
        def decorator(fn):
            self.on(eventname=eventname,function=fn,once=once)
        if function is not None: return decorator(function)
        else: return decorator

def TokenReader(filename: str) -> str:
    """
    Read the first line of any file and return it, making it useful for securing token.
    """
    try:
        with open(filename, 'r') as file:
            tokenline = file.readline()
            return tokenline
    except FileNotFoundError:
        raise FileNotFoundError(f"File '{filename}' not found.")
    except Exception as e:
        raise ReferenceError(f"File '{filename}' cannot be read by TokenReader() with error {e}.")

# I keep this to support old version
@deprecated('This function is deprecated, use EmbedBuilder instead!')
def Embed(**kwargs):
    content = {}
    content.update(kwargs)
    content['type'] = 'rich'
    return content

# Types
from dispy.types.message import Message
from dispy.types.interaction import Interaction
from dispy.types.user import User
from dispy.types.embed import EmbedBuilder
from dispy.types.reaction import ReactionAdd, ReactionRemove, ReactionRemoveAll, ReactionRemoveEmoji
typesFunc = ['Message','User', 'Interaction', 'ReactionAdd', 'ReactionRemove', 'ReactionRemoveAll', 'ReactionRemoveEmoji']

# END
__all__ = ['Bot','EmbedBuilder','Embed','TokenReader'] + typesFunc


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dispy.modules.dictwrapper import DictWrapper
from dispy.types.variable import Snowflake, Timestamp
from typing import List, Dict, Any

class Attachment(DictWrapper):
    id: Snowflake
    filename: str
    title: str
    description: str
    content_type: str
    size: int
    url: str
    proxy_url: str
    height: int
    width: int
    ephemeral: bool
    duration_secs: float
    waveform: str
    flags: int


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# When check, using ON and ONCE will enforce using custom args
intents = {
   "-1": [
      "READY",
      "RESUMED",
      "VOICE_SERVER_UPDATE",
      "USER_UPDATE",
      "INTERACTION_CREATE",
      "ALL"
   ],
   "0": [
      "GUILD_CREATE",
      "GUILD_UPDATE",
      "GUILD_DELETE",
      "GUILD_ROLE_CREATE",
      "GUILD_ROLE_UPDATE",
      "GUILD_ROLE_DELETE",
      "CHANNEL_CREATE",
      "CHANNEL_UPDATE",
      "CHANNEL_DELETE",
      "CHANNEL_PINS_UPDATE",
      "THREAD_CREATE",
      "THREAD_UPDATE",
      "THREAD_DELETE",
      "THREAD_LIST_SYNC",
      "THREAD_MEMBER_UPDATE",
      "THREAD_MEMBERS_UPDATE",
      "STAGE_INSTANCE_CREATE",
      "STAGE_INSTANCE_UPDATE",
      "STAGE_INSTANCE_DELETE",
      "MESSAGE_CREATE",
      "MESSAGE_UPDATE",
      "MESSAGE_DELETE",
      "MESSAGE_DELETE_BULK",
      "ALL"
   ],
   "1": [
      "GUILD_MEMBER_ADD",
      "GUILD_MEMBER_UPDATE",
      "GUILD_MEMBER_REMOVE",
      "THREAD_MEMBERS_UPDATE",
      "ALL"
   ],
   "2": [
      "GUILD_AUDIT_LOG_ENTRY_CREATE",
      "GUILD_BAN_ADD",
      "GUILD_BAN_REMOVE",
      "ALL"
   ],
   "3": [
      "GUILD_EMOJIS_UPDATE",
      "GUILD_STICKERS_UPDATE",
      "GUILD_SOUNDBOARD_SOUND_CREATE",
      "GUILD_SOUNDBOARD_SOUND_UPDATE",
      "GUILD_SOUNDBOARD_SOUND_DELETE",
      "GUILD_SOUNDBOARD_SOUNDS_UPDATE",
      "ALL"
   ],
   "4": [
      "GUILD_INTEGRATIONS_UPDATE",
      "INTEGRATION_CREATE",
      "INTEGRATION_UPDATE",
      "INTEGRATION_DELETE",
      "ALL"
   ],
   "5": [
      "WEBHOOKS_UPDATE",
      "ALL"
   ],
   "6": [
      "INVITE_CREATE",
      "INVITE_DELETE",
      "ALL"
   ],
   "7": [
      "VOICE_CHANNEL_EFFECT_SEND",
      "VOICE_STATE_UPDATE",
      "ALL"
   ],
   "8": [
      "PRESENCE_UPDATE",
      "ALL"
   ],
   "9": [
      "MESSAGE_CREATE",
      "MESSAGE_UPDATE",
      "MESSAGE_DELETE",
      "MESSAGE_DELETE_BULK",
      "ALL"
   ],
   "10": [
      "MESSAGE_REACTION_ADD",
      "MESSAGE_REACTION_REMOVE",
      "MESSAGE_REACTION_REMOVE_ALL",
      "MESSAGE_REACTION_REMOVE_EMOJI",
      "ALL"
   ],
   "11": [
      "TYPING_START",
      "ALL"
   ],
   "12": [ 
      "DIRECT_MESSAGE_CREATE",
      "DIRECT_MESSAGE_UPDATE",
      "DIRECT_MESSAGE_DELETE",
      "DIRECT_CHANNEL_PINS_UPDATE",
      "ALL"
   ],
   "13": [
      "DIRECT_MESSAGE_REACTION_ADD",
      "DIRECT_MESSAGE_REACTION_REMOVE",
      "DIRECT_MESSAGE_REACTION_REMOVE_ALL",
      "DIRECT_MESSAGE_REACTION_REMOVE_EMOJI",
      "ALL"
   ],
   "14": [
      "DIRECT_TYPING_START",
      "ALL"
   ],
   "15": [
      "MESSAGE_CREATE",
      "MESSAGE_UPDATE",
      "MESSAGE_DELETE",
      "MESSAGE_DELETE_BULK",
      "MESSAGE_POLL_VOTE_ADD",
      "MESSAGE_POLL_VOTE_REMOVE",
      "ALL"
   ],
   "16": [
      "GUILD_SCHEDULED_EVENT_CREATE",
      "GUILD_SCHEDULED_EVENT_UPDATE",
      "GUILD_SCHEDULED_EVENT_DELETE",
      "GUILD_SCHEDULED_EVENT_USER_ADD",
      "GUILD_SCHEDULED_EVENT_USER_REMOVE",
      "ALL"
   ],
   "20": [
      "AUTO_MODERATION_RULE_CREATE",
      "AUTO_MODERATION_RULE_UPDATE",
      "AUTO_MODERATION_RULE_DELETE",
      "ALL"
   ],
   "21": [
      "AUTO_MODERATION_ACTION_EXECUTION",
      "ALL"
   ],
   "24": [
      "MESSAGE_POLL_VOTE_ADD",
      "MESSAGE_POLL_VOTE_REMOVE",
      "ALL"
   ],
   "25": [
      "DIRECT_MESSAGE_POLL_VOTE_ADD",
      "DIRECT_MESSAGE_POLL_VOTE_REMOVE",
      "ALL"
   ]
}


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dispy.modules.dictwrapper import DictWrapper
from dispy.types.variable import Snowflake, Timestamp
from typing import List, Dict, Any

from dispy.types.emoji import Emoji
from dispy.types.channel import Channel


class ComponentDefaultValue(DictWrapper):
    id: Snowflake
    type: str


class ComponentOption(DictWrapper):
    label: str
    value: str
    description: str
    emoji: Emoji
    default: bool


class ComponentObject(DictWrapper):
    type: int
    style: int
    label: str
    emoji: Emoji
    custom_id: str
    sku_id: Snowflake
    url: str
    value: str
    required: bool
    disabled: bool
    options: List[ComponentOption]
    channel_types: List[Channel]
    placeholder: str
    default_values: List[ComponentDefaultValue]
    min_values: int
    max_values: int


class Component(DictWrapper):
    type: int
    components: List[ComponentObject]


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dispy.modules.dictwrapper import DictWrapper
from dispy.types.variable import Snowflake, Timestamp
from typing import List, Dict, Any, TypedDict, Literal
import re

class EmbedField(DictWrapper):
    name: str
    value: str
    inline: bool

class EmbedAuthor(DictWrapper):
    name: str
    url: str
    icon_url: str
    proxy_icon_url: str

class EmbedProvider(DictWrapper):
    name: str
    url: str

class EmbedFooter(DictWrapper):
    text: str
    icon_url: str
    proxy_icon_url: str

class Embed(DictWrapper):
    title: str
    type: str
    description: str
    url: str
    timestamp: Timestamp
    color: int
    footer: EmbedFooter
    image: Any
    thumbnail: Any
    video: Any
    provider: EmbedProvider
    author: EmbedAuthor
    fields: List[EmbedField]

class EmbedBuilder:
    """
    You can build an embed like in discord.js, template to get started:
    ```py
    embed = (EmbedBuilder()
        .setTitle('Title')
        .setDescription('Very interesting title')
        .setColor('red')
    )
    
    msg.send(embeds=embed)
    ```
    """
    def __init__(self):
        self.args = {}
        self.fields = []
    def get(self):
        content = {}
        content.update(self.args)
        if len(self.fields) > 0:
            content.update({'fields': self.fields})
        content['type'] = 'rich'
        return content
    
    def setTitle(self, title: str):
        self.args['title'] = title
        return self    
    def setDescription(self, title: str):
        self.args['description'] = title
        return self    
    
    class authorObject(TypedDict):
        name: str
        icon_url: str
        proxy_icon_url: str
        url: str
    def setAuthor(self, author: dict | authorObject):
        self.args['author'] = author
        return self
    
    def setUrl(self, url: str):
        self.args['url'] = url
        return self
    def setTimestamp(self, timestamp: Timestamp):
        """
        Take ISO8601, may support more in the future.
        """
        self.args['timestamp'] = timestamp
        return self
    
    _colors = Literal['blue','violet','pink','red','orange','yellow','green','black','brown','gray','white']
    global _color_palette
    _color_palette = {
        'blue': 0x3983F2,
        'violet': 0x9239F2,
        'pink': 0xF239D0,
        'red': 0xF23939,
        'orange': 0xF29C39,
        'yellow': 0xF2F239,
        'green': 0x39F23F,
        'black': 0x000000,
        'brown': 0xA05601,
        'gray': 0xA9A9A9,
        'white': 0xFFFFFF
    }
    def setColor(self, color: str | int | _colors):
        """
        Support hexcode, integer color and custom color (Red, blue, etc...)
        """
        color = color.lower()
        if isinstance(color, str) and color in list(_color_palette.keys()):
            self.args['color'] = _color_palette[color]
        elif isinstance(color, str):
            pattern = re.compile(r'^#?([0-9A-Fa-f]{6})$')
            match = pattern.match(color)
            if not match:
                raise ValueError("Invalid hex color code. Please provide a valid hex code in the format '#RRGGBB' or 'RRGGBB'.")
            self.args['color'] = int(match.group(1), 16)
            
        elif isinstance(color, int):
            if not (0x000000 <= color <= 0xFFFFFF):
                raise TypeError(f'{color} isn\'t valid. Try hexcode.')
            self.args['color'] = color
        else:
            raise TypeError('Type not compatible.')
        return self

    class FooterObject(TypedDict):
        text: str
        icon_url: str
        proxy_icon_url: str
    def setFooter(self, footer: dict | FooterObject):
        self.args['footer'] = footer
        return self
    
    class ImageObject(TypedDict):
        url: str
        proxy_url: str
        height: int
        width: int
    def setImage(self, image: dict | ImageObject):
        self.args['image'] = image
        return self
    
    class ThumbnailObject(TypedDict):
        url: str
        proxy_url: str
        height: int
        width: int
    def setThumbnail(self, thumbnail: dict | ThumbnailObject):
        self.args['thumbnail'] = thumbnail
        return self
    
    class VideoObject(TypedDict):
        url: str
        proxy_url: str
        height: int
        width: int
    def setVideo(self, video: dict | VideoObject):
        self.args['video'] = video
        return self

    class ProviderObject(TypedDict):
        name: str
        url: str
    def setProvider(self, provider: dict | ProviderObject):
        self.args['provider'] = provider
        return self
    
    # Fields
    class Field(TypedDict):
        name: str
        value: str
        inline: bool

    def setFields(self, *fields: List[Field] | Field):
        """
        Set one or multiple fields.
        ⚠️ Will remove fields defined before.
        """
        fields = [item for arg in fields for item in (arg if isinstance(arg, list) else [arg])]
        self.fields = fields
        return self
    
    def setField(self, name: str, value: str, inline: bool = None):
        """
        Set one field.
        ⚠️ Will remove fields defined before.
        """
        field = {'name': name, 'value': value}
        if inline is not None:
            field['inline'] = inline
        self.fields = [field]
        return self
    
    def addField(self, name: str, value: str, inline: bool = None):
        """
        Add a field.
        """
        field = {'name': name, 'value': value}
        if inline is not None:
            field['inline'] = inline
        self.fields.append(field)
        return self
    
    def addFields(self, *fields: List[Field] | Field):
        """
        Add one or multiple fields.
        """
        fields = [item for arg in fields for item in (arg if isinstance(arg, list) else [arg])]
        self.fields.extend(fields)
        return self


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

errors = {
    "traceback": "Traceback (most recent call last):",
    "no_traceback": "No traceback available.",
    "file": "File \"{filename}\", line {line}, in {name}",
    "event_invalid": "The event {event} doesn't exist",
    "noargs": "Event {eventname} connected to {function_name}() doesn't give any arguments. Remove them!",
    "invalidtype_args": "Argument '{name}' of {function_name}() need to be of type {needed_type}. The correct way:\ndef {function_name}({arguments}): ...",
    "extra_args": "{function_name}() has '{name}' which is invalid. The correct way:\ndef {function_name}({arguments}): ...",
    "missing_args": "{function_name}() missing argument '{name}'. The correct way:\ndef {function_name}({arguments}): ...",
    "bot_is_already_running": "The bot is already running",
    "bot_is_running": "The bot is currently active and you cannot change anything",
    "request_failed": "The HTTP request code is {code} with error \"{error}\"",
    "dispy_request_error": "An error occured with dispy during HTTP request: {error}",
    "getting_invalid": "You've tried to get() an invalid result, error of the result: {error}",
    "invalid_arguments": "You didn't give the argument {arg}!"
}


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dispy.modules.dictwrapper import DictWrapper
from dispy.types.variable import Snowflake, Timestamp
from dispy.modules.rest_api import result
from typing import List, Dict, Any
import asyncio

from dispy.types.message import Message
from dispy.types.guild import Member
from dispy.types.guild import Guild, Member
from dispy.types.channel import Channel
from dispy.types.message import ResolvedData
from dispy.types.emoji import Emoji
from dispy.types.entitlement import Entitlement
from dispy.types.user import User

class CommandInteractionDataOptions(DictWrapper):
    name: str
    type: int
    value: str | int | float | bool
    options: List['CommandInteractionDataOptions']
    focused: bool

class SelectOption(DictWrapper):
    label: str
    value: str
    description: str
    emoji: Emoji
    default: bool

class InteractionData(DictWrapper): # Unfortunatly, interaction data can be different so i put every possibilities in there
    id: Snowflake
    custom_id: str
    name: str
    type: int
    resolved: ResolvedData
    component_type: int
    components: Any
    values: List[SelectOption]
    options: List[CommandInteractionDataOptions]
    guild_id: Snowflake
    target_id: Snowflake

class Interaction(DictWrapper):
    id: Snowflake
    application_id: Snowflake
    version: int
    type: int
    token: str
    message: Message
    member: Member
    locale: str
    guild_locale: str
    guild_id: Snowflake
    guild: Guild
    channel_id: Snowflake
    channel: Channel
    entitlements: list
    entitlement_sku_ids: list
    data: InteractionData
    member: Member
    user: User
    entitlements: List[Entitlement]
    authorizing_integration_owners: Dict[str,Any]
    context: int
    app_permissions: str
    _api = None

    def reply(self,content=None,**kwargs) -> result["Message"]:
        """
        Reply to the message.
        """
        future = self._api._loop.create_future()
        
        async def _asynchronous(content, **kwargs):
            payload = {}
            #    "message_reference": {
            #        "channel_id": self.channel_id,
            #        "message_id": self.id,
            #        "guild_id": self.guild_id,
            #        "type": 0
            #    },
            #    "type": 19
            #}
    
            # Embed
            embeds = kwargs.get('embeds',None)
            if embeds is not None:
                if not isinstance(embeds, list):
                    embeds = [embeds]
                kwargs['embeds'] = embeds
            
            if kwargs:
                payload.update(kwargs)
            if content:
                payload.update({"content": content})
            
            result = await self._api.__request__('post', f'webhooks/{self.application_id}/{self.token}', payload) # no_traceback
            future.set_result(result)
        
        asyncio.run_coroutine_threadsafe(_asynchronous(content, **kwargs), self._api._loop)
        return result[Message](future,self._api,Message)


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

#"-1": [
#   "READY",
#   "RESUMED",
#   "VOICE_SERVER_UPDATE",
#   "USER_UPDATE",
#   "INTERACTION_CREATE",
#   "ALL"
#],

class intents_variable:
    def __init__(self,intents) -> None:
        intents_list = intents.copy()
        del intents_list['-1']
        self.__intents_list__ = intents_list

        self.intents = {key for nested_dict in intents.values() for key in nested_dict}
        self.direct_intents = intents["12"] + intents["13"] + intents["14"] + intents["25"]
        self.direct_intents_opposed = [intents["0"][9]] + intents["9"][:3] + intents["10"] + intents["11"] + intents["24"]
    def get_intents(self,eventname) -> list:
        parents = []
        for parent_id, events in self.__intents_list__.items():
            if eventname in events:
                parents.append(parent_id)
        return parents
    def get_child(self,id):
        return list(self.__intents_list__.values())[id]


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dispy.modules.dictwrapper import DictWrapper
from dispy.types.variable import Snowflake, Timestamp
from typing import List, Dict, Any

from dispy.types.guild import Member

class ThreadMember(DictWrapper):
    id: Snowflake
    user_id: Snowflake
    join_timestamp: Timestamp
    flags: int
    member: Member

class ThreadMetadata(DictWrapper):
    archived: bool
    auto_archive_duration: int
    archive_timestamp: Timestamp
    locked: bool
    invitable: bool
    create_timestamp: Timestamp

class ForumTag(DictWrapper):
    id: Snowflake
    name: str
    moderated: bool
    emoji_id: Snowflake
    emoji_name: str

class DefaultReaction(DictWrapper):
    emoji_id: Snowflake
    emoji_name: str


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import get_type_hints, Any, Dict, List, Union, _GenericAlias
import inspect
import json


# This variable is only On for debugging purposes and to see if there is missing args, need to be Off for normal uses.
debug = True

class DictWrapper:
    """
    DictWrapper is a tool used by Dispy made to replace TypedDict, basemodel from pydantic and dataclass. It is made to have only the pros of theses three.

    ### Functionalities
    - Take Dict as input (Pydantic & dataclass)
    - Support self-referencing (TypedDict)
    - Get by obj (Pydantic & dataclass)
    """
    #This function is needed to .dumps a text inside DictWrapper, but i removed it because it is monkey patching
    #and this is a programming crime. If you know how or want to fix this problem, contribute to dispy!
    #
    #def _jsonSupport():
    #    original_default = json.JSONEncoder.default
    #    original_decoder = json.JSONDecoder().object_hook
    #    def default(self, obj):
    #        if isinstance(obj, DictWrapper):
    #            return {'type': 'DictWrapper', 'name': obj.to_dict()}
    #        return original_default(self, obj)
    #    def object_hook(obj):
    #        if 'type' in obj and obj['type'] == 'DictWrapper':
    #            return DictWrapper.from_dict(obj['name'])
    #        return obj if original_decoder is None else original_decoder(obj)
    #    json.JSONEncoder.default = default
    #    json._default_decoder = json.JSONDecoder(object_hook=object_hook)
    #_jsonSupport()

    _dictwrapper = True
    _api = None
    _types = None

    def __init__(self, **kwargs):
        """
        This class uses DictWrapper!
        """
        self._types = get_type_hints(self.__class__)

        for key, value in kwargs.items():
            if key not in self._types and key != '_api' and debug:
                raise KeyError(f"'{key}' was given but isn't defined in the DictWrapper '{self.__class__.__name__}'.")

            if '_api' in kwargs:
                self._api = kwargs['_api']

            if isinstance(value, dict):
                if not isinstance(self._types[key], _GenericAlias) and issubclass(self._types[key], DictWrapper) and self._api is not None:
                    value = self._types[key](_api=self._api, **value)
                elif inspect.isclass(self._types[key]):
                    value = self._types[key](**value)
                else:
                    value = value
            setattr(self, key, value)

    def __getattr__(self, name):
        if name in self._types:
            return None
        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{name}'")
    
    def __getitem__(self, item):
        if hasattr(self, item):
            value = getattr(self, item)
            if isinstance(value, DictWrapper):
                return value.to_dict()
            return value
        raise KeyError(f"'{item}' not found in '{self.__class__.__name__}'")
    
    def to_dict(self):
        result = {}
        for key in self._types:
            value = getattr(self, key, None)
            if isinstance(value, DictWrapper):
                result[key] = value.to_dict()
            else:
                result[key] = value
        return result
    
    def __iter__(self):
        return iter(self.to_dict().items())
    
    def __repr__(self):
        return repr(self.to_dict())
    
    def __str__(self):
        return str(self.to_dict())


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dispy.modules.dictwrapper import DictWrapper
from dispy.types.variable import Snowflake, Timestamp
from dispy.modules.rest_api import result
import asyncio
from typing import List, Dict, Any

from dispy.types.user import User
from dispy.types.role import Role, RoleSubscriptionData
from dispy.types.embed import Embed, EmbedBuilder
from dispy.types.channel import ChannelMention, Channel
from dispy.types.attachment import Attachment
from dispy.types.reaction import Reaction
from dispy.types.application import Application
from dispy.types.guild import Member
from dispy.types.component import Component
from dispy.types.stickers import StickerItem, Sticker
from dispy.types.poll import Poll

class MessageActivity(DictWrapper):
    type: int
    party_id: str

class MessageReference(DictWrapper):
    type: int
    message_id: Snowflake
    channel_id: Snowflake
    guild_id: Snowflake
    fail_if_not_exists: bool

class MessageSnapshot(DictWrapper):
    message: 'Message'

class MessageInteractionMetadata(DictWrapper):
    id: Snowflake
    type: int
    name: str
    user: User
    authorizing_integration_owners: Dict[Any,Any]
    original_response_message_id: Snowflake
    target_user: User
    target_message_id: Snowflake
    command_type: int

class MessageInteraction(DictWrapper):
    id: Snowflake
    type: int
    name: str
    user: User
    member: Member

class ResolvedData(DictWrapper):
    users: List[User]
    members: List[Member]
    roles: List[Role]
    channels: List[Channel]
    messages: List["Message"]
    attachments: List[Attachment]

class MessageCall(DictWrapper):
    participants: List[Snowflake]
    ended_timestamp: Timestamp

class Message(DictWrapper):
    id: Snowflake
    channel_id: Snowflake
    author: User
    content: str
    timestamp: Timestamp
    edited_timestamp: Timestamp
    tts: bool
    mention_everyone: bool
    mentions: List[User]
    mention_roles: List[Role]
    mention_channels: List[ChannelMention]
    attachments: List[Attachment]
    embeds: List[Embed]
    reactions: List[Reaction]
    nonce: bool | int
    pinned: bool
    webhook_id: Snowflake
    type: int
    activity: MessageActivity
    application: Application
    application_id: Snowflake
    channel_type: int
    member: Member
    guild_id: Snowflake
    flags: int
    message_reference: MessageReference
    message_snapshots: List[MessageSnapshot]
    referenced_message: 'Message'
    interaction_metadata: MessageInteractionMetadata
    interaction: MessageInteraction
    thread: Channel
    components: List[Component]
    sticker_items: List[StickerItem]
    stickers: List[Sticker] # Useless lol, this shit is deprecated
    position: int
    role_subscription_data: RoleSubscriptionData
    resolved: ResolvedData
    poll: Poll
    call: MessageCall
    _api = None

    # Message
    def reply(self,content=None, embeds=None, **kwargs) -> result["Message"]:
        """
        Reply to the message.
        """
        future = self._api._loop.create_future()
        
        async def _asynchronous(embeds):
            payload = {
                "message_reference": {
                    "channel_id": self.channel_id,
                    "message_id": self.id,
                    "guild_id": self.guild_id,
                    "type": 0
                },
                "type": 19
            }

            # Embed
            if isinstance(embeds, list):
                embeds = [embed.get() if isinstance(embed, EmbedBuilder) else embed for embed in embeds]
            else:
                if isinstance(embeds, EmbedBuilder):
                    embeds = embeds.get()
            if embeds is not None:
                if not isinstance(embeds, list):
                    embeds = [embeds]
                payload.update({"embeds": embeds})
            
            # Reste
            if kwargs:
                payload.update(kwargs)
            if content:
                payload.update({"content": content})
            
            result = await self._api.__request__('post', f'channels/{self.channel_id}/messages', payload) # no_traceback
            future.set_result(result)
        
        asyncio.run_coroutine_threadsafe(_asynchronous(embeds), self._api._loop)
        return result[Message](future,self._api,Message)
    
    def send(self,content=None, embeds=None, reply_to: 'Message' = None, **kwargs) -> result["Message"]:
        """
        Send a message in the same channel as the message.
        """
        future = self._api._loop.create_future()
        
        async def _asynchronous(embeds):
            payload = {}

            # Reply
            if reply_to != None:
                payload.update({
                    "message_reference": {
                        "channel_id": reply_to.channel_id,
                        "message_id": reply_to.id,
                        "guild_id": reply_to.guild_id,
                        "type": 0
                    },
                    "type": 19
                })
    
            # Embed
            if isinstance(embeds, list):
                embeds = [embed.get() if isinstance(embed, EmbedBuilder) else embed for embed in embeds]
            else:
                if isinstance(embeds, EmbedBuilder):
                    embeds = embeds.get()
            if embeds is not None:
                if not isinstance(embeds, list):
                    embeds = [embeds]
                payload.update({"embeds": embeds})
            
            # Reste
            if kwargs:
                payload.update(kwargs)
            if content:
                payload.update({"content": content})
            
            result = await self._api.__request__('post', f'channels/{self.channel_id}/messages', payload) # no_traceback
            future.set_result(result)
        
        asyncio.run_coroutine_threadsafe(_asynchronous(embeds), self._api._loop)
        return result[Message](future,self._api,Message)
    
    def delete(self) -> result[None]:
        """
        Delete the message.
        """
        future = self._api._loop.create_future()
        
        async def _asynchronous(channel_id,message_id):
            result = await self._api.__request__('delete', f'channels/{channel_id}/messages/{message_id}') # no_traceback
            future.set_result(result)
        
        asyncio.run_coroutine_threadsafe(_asynchronous(self.channel_id, self.id), self._api._loop)
        return result[None](future,self._api,None)
    
    def edit(self,content=None, embeds=None, **kwargs):
        """
        Edit the message.
        You need to be the author of it to edit it.
        """
        future = self._api._loop.create_future()
        
        async def _asynchronous(embeds):
            payload = {}
        
            # Embed
            if isinstance(embeds, list):
                embeds = [embed.get() if isinstance(embed, EmbedBuilder) else embed for embed in embeds]
            else:
                if isinstance(embeds, EmbedBuilder):
                    embeds = embeds.get()
            if embeds is not None:
                if not isinstance(embeds, list):
                    embeds = [embeds]
                payload.update({"embeds": embeds})
            
            # Reste
            if kwargs:
                payload.update(kwargs)
            if content:
                payload.update({"content": content})
            
            result = await self._api.__request__('patch', f'channels/{self.channel_id}/messages/{self.id}', payload) # no_traceback
            future.set_result(result)
        
        asyncio.run_coroutine_threadsafe(_asynchronous(embeds), self._api._loop)
        return result[Message](future,self._api,Message)


from setuptools import setup, find_packages

setup(
    name='dispy-bot',
    version='0.1.0.6',
    description='A python-coded discord bot library.',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='James French',
    author_email='jamesfrench.contact@gmail.com',
    url='https://github.com/JamesMinoucha/Dispy',
    packages=find_packages(),
    install_requires=[
        "aiohttp>=3.10.10",
        "pydantic>=2.10.2",
        "pydantic_core>=2.27.1",
        "websocket_client>=1.8.0",
    ],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.12',
)


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

class Snowflake(str):
    def __new__(cls, value):
        return str.__new__(cls, value)

class Timestamp(str):
    def __new__(cls, value):
        return str.__new__(cls, value)
    
class Invalid(str):
    def __new__(cls, value):
        return str.__new__(cls, value)
    
class Null(str):
    def __init__(self):
        pass


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dispy.modules.dictwrapper import DictWrapper
from dispy.types.variable import Snowflake, Timestamp
from typing import List, Dict, Any

class AvatarDecorationData(DictWrapper):
    asset: str
    sku_id: Snowflake
    expires_at: int

class User(DictWrapper):
    id: Snowflake
    username: str
    discriminator: str
    display_name: str
    global_name: str
    avatar: str
    clan: Any
    bot: bool
    system: bool # This is useless x)
    mfa_enabled: bool
    verified: bool
    email: str
    locale: str
    flags: int
    banner: str
    banner_color: int
    accent_color: int
    premium_type: int
    public_flags: int
    primary_guild: str
    avatar_decoration_data: AvatarDecorationData
    _api = None


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dispy.modules.dictwrapper import DictWrapper
from dispy.types.variable import Snowflake, Timestamp
from typing import List, Dict, Any

from dispy.types.role import Role
from dispy.types.user import User

class Emoji(DictWrapper):
    id: Snowflake
    name: str
    roles: List[Role]
    user: User
    require_colons: bool
    managed: bool
    animated: bool
    available: bool


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

class dict_to_obj(dict):
    def __getattr__(self, attr):
        value = self.get(attr)
        if isinstance(value, dict):
            return dict_to_obj(value)
        return value
    
def format_list(elements):
    if not elements:
        return ""
    elif len(elements) == 1:
        return elements[0]
    elif len(elements) == 2:
        return " and ".join(elements)
    else:
        return ", ".join(elements[:-1]) + " and " + elements[-1]


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import aiohttp
import asyncio
import json
import threading
from dispy.modules import dict_to_obj
from dispy.types.embed import EmbedBuilder
from dispy.types.variable import Invalid
from typing import Generic, TypeVar, Type

# 888888ba  oo                               
# 88    `8b ``                               
# 88     88 dP .d8888b.    88d888b. dP    dP 
# 88     88 88 Y8ooooo.    88'  `88 88    88 
# 88    .8P 88       88 ,, 88.  .88 88.  .88 
# 8888888P  dP `88888P' 88 88Y888P' `8888P88 
#                          88            .88 
#                          dP        d8888P  

# Developed by ✯James French✯ with ❤
# Licensed with GPLv3

T = TypeVar('T')

class result(Generic[T]):
    """
    Represent the response of the request you've done.
    Use .get() to get variables, this will pause the execution of your code until the response is given.
    """
    def __init__(self, future: asyncio.Future[T], api, cls: Type[T] = None):
        self.future = future
        self.api = api
        self.loop = self.api._loop
        self._cls = cls
    
    def get(self) -> T:
        """
        Will block the code execution until response is given.
        """
        if not isinstance(self.future,Invalid):
            async def asynchronous() -> T:
                return await self.future
            
            future_result = asyncio.run_coroutine_threadsafe(asynchronous(), self.loop)
            result = future_result.result(timeout=7)
        else:
            self.api._error.summon("getting_invalid",stop=False,error=self.future)
            return None
        
        if isinstance(result,Invalid):
            self.api._error.summon("getting_invalid",stop=False,error=result)
            return None
        else:
            return self._cls(**result, _api=self.api) if hasattr(self._cls,'_api') else self._cls(**result)




from dispy.types.message import Message

class __internal__(Generic[T]):
    def  __init__(self,token,error_handler) -> None:
        self._token = token
        self._header = {
            'authorization': f'Bot {self._token}',
            'content-type': 'application/json'
        }
        self._loop = asyncio.new_event_loop()
        self._base_url = 'https://discord.com/api/v10/'
        self._api = self
        self._error = error_handler
        threading.Thread(target=self.run_loop, daemon=True).start()
    def run_loop(self):
        asyncio.set_event_loop(self._loop)
        self._loop.run_forever() # no_traceback

    async def __request__(self,function,path,payload=None):
        args = {}
        if payload:
            args['json'] = payload
        async with aiohttp.ClientSession() as session:
            try:
                async with getattr(session, function)(f'{self._base_url}{path}', headers=self._header, **args) as response:
                    if response.status not in [200, 204]:
                        error = json.loads(await response.text())
                        self._error.summon("request_failed",stop=False,code=response.status,error=error["message"])
                        return Invalid(self._error.get("request_failed",code=response.status,error=error["message"]))
                    else:
                        if response.status != 204:
                            response_data = await response.json()
                            return dict_to_obj(response_data)
                        else:
                            return None
            except Exception as err:
                self._error.summon("dispy_request_error",stop=False,error=err)

    #--------------------------------------------------------------------------------------#
    #                                       Requests                                       #
    #--------------------------------------------------------------------------------------#
    def create_message(self,content=None, channel_id=None, embeds=None, reply_to: Message = None, **kwargs) -> result[Message]:
        
        """
        Send a message in a specific channel.
        """
        future = self._api._loop.create_future()

        async def _asynchronous(embeds):
            payload = {}
    
            # Reply
            if reply_to != None:
                payload.update({
                    "message_reference": {
                        "channel_id": reply_to.channel_id,
                        "message_id": reply_to.id,
                        "guild_id": reply_to.guild_id,
                        "type": 0
                    },
                    "type": 19
                })
    
            # Embed
            if isinstance(embeds, list):
                embeds = [embed.get() if isinstance(embed, EmbedBuilder) else embed for embed in embeds]
            else:
                if isinstance(embeds, EmbedBuilder):
                    embeds = embeds.get()
            if embeds is not None:
                if not isinstance(embeds, list):
                    embeds = [embeds]
                payload.update({"embeds": embeds})
            
            # Reste
            if kwargs:
                payload.update(kwargs)
            if content:
                payload.update({"content": content})
            
            result = await self._api.__request__('post', f'channels/{channel_id}/messages', payload) # no_traceback
            future.set_result(result)
        
        asyncio.run_coroutine_threadsafe(_asynchronous(embeds), self._api._loop)
        return result[Message](future,self._api,Message)

    def delete_message(self,channel_id,message_id) -> result[None]:
        """
        Delete a specific message.
        """
        future = self._api._loop.create_future()

        async def _asynchronous(channel_id,message_id):
            result = await self._api.__request__('delete', f'channels/{channel_id}/messages/{message_id}') # no_traceback
            future.set_result(result)
        
        asyncio.run_coroutine_threadsafe(_asynchronous(channel_id, message_id), self._api._loop)
        return result[None](future,self._api,None)
    
    def edit_message(self,content=None, channel_id=None, message_id=None, embeds=None, **kwargs) -> result[Message]:
        
        """
        Edit a specific message in a specific channel.
        You need to be the author of the message to edit it.
        """
        future = self._api._loop.create_future()
    
        async def _asynchronous(embeds):
            payload = {}

            # Embed
            if isinstance(embeds, list):
                embeds = [embed.get() if isinstance(embed, EmbedBuilder) else embed for embed in embeds]
            else:
                if isinstance(embeds, EmbedBuilder):
                    embeds = embeds.get()
            if embeds is not None:
                if not isinstance(embeds, list):
                    embeds = [embeds]
                payload.update({"embeds": embeds})
            
            # Reste
            if kwargs:
                payload.update(kwargs)
            if content:
                payload.update({"content": content})
            
            result = await self._api.__request__('patch', f'channels/{channel_id}/messages/{message_id}', payload) # no_traceback
            future.set_result(result)
        
        asyncio.run_coroutine_threadsafe(_asynchronous(embeds), self._api._loop)
        return result[Message](future,self._api,Message)


# Dispy - Python Discord API library for discord bots.
# Copyright (C) 2024  James French
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import traceback
import sys
import re
import os
import asyncio
import threading
from dispy.data import errors

# Custom error handling for dispy

class error:
    def __init__(self):
        self.errors = errors
        self.asyncio_path = os.path.dirname(asyncio.__file__)
        self.threading_path = os.path.dirname(threading.__file__)
    def summon(self,error_name,stop=True,**kwargs):
        """
        Custom error handling, do not use if you don't know what your doing.
        """
        # Print the traceback
        stack = traceback.extract_stack()[:-2]
        filtered_stack = [frame for frame in stack if not re.compile(r'#\s*no_traceback\s*$').search(frame.line.replace(' ',''))]
        filtered_stack = [frame for frame in filtered_stack if not frame.filename.startswith(self.asyncio_path)]
        filtered_stack = [frame for frame in filtered_stack if not frame.filename.startswith(self.threading_path)]

        if len(filtered_stack) > 0:
            print('\033[93m' + self.errors['traceback'] + '\033[0m')
            for frame in filtered_stack:
                print(f"  {self.errors['file'].format(filename=frame.filename,line=frame.lineno,name=frame.name)}")
                print(f"    {frame.line}")
        else:
            print('\033[93m' + self.errors['no_traceback'] + '\033[0m')

        error = self.errors[error_name].format(**kwargs)
        print(f'\033[31m{error}\033[0m')    
    
        # Exit the program
        if stop: sys.exit()
        else: return error
    def get(self,error_name,**kwargs):
        return self.errors[error_name].format(**kwargs)